#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// standard includes
#include <stdexcept>  // runtime_error 
#include <iostream>   // cout
#include <fstream>
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>  // ostrstream, istrstream
#endif

// the general include for eo

#include <eo>
#include <ga.h>

// my stuff 
#include "operators.h"
#include "fitness.h"
#include "check.h"

#include "gui/gui.h"


using namespace std;


//-----------------------------------------------------------------------------
void main_function(int argc, char **argv)
{


  // PARAMETRES
   unsigned int SEED;      // seed for random number generator
   unsigned int POP_SIZE;  // Size of population
   unsigned int MAX_GEN;  // Maximum number of generation before STOP
   double CROSS_RATE;      // Crossover rate
   double P_MUT_PER_BIT; // probability of bit-flip mutation
   double MUT_RATE;        // mutation rate
   double P_MUT1, P_MUT2;
   double elita; 

//-----------------------------------------------------------------------------
// instead of having all values of useful parameters as constants, read them:
// either on the command line (--option=value or -o=value)
//     or in a parameter file (same syntax, order independent, 
//                             # = usual comment character 


   eoParser parser(argc, argv);

    eoValueParam<int> seedParam(time(0), "seed", "Random number seed", 'S');
    parser.processParam( seedParam );
    SEED = seedParam.value();

    
    eoValueParam<unsigned int> popSizeParam(50, "popSize", "Population size",'P');
    parser.processParam( popSizeParam, "Evolution engine" );
    POP_SIZE = popSizeParam.value();

    eoValueParam<unsigned int> maxGenParam(10000, "maxGen", "Maximum number of generations",'G');
    parser.processParam( maxGenParam, "Stopping criterion" );
    MAX_GEN = maxGenParam.value();

    eoValueParam<double> pCrossParam(0.3, "pCross", "Probability of Crossover", 'C'); 
    parser.processParam( pCrossParam, "Genetic Operators" );
    CROSS_RATE = pCrossParam.value();

    eoValueParam<double> pMutParam(0.1, "pMut", "Probability of Mutation", 'M');
    parser.processParam( pMutParam, "Genetic Operators" );
    MUT_RATE = pMutParam.value();

    
    eoValueParam<double> pMutPerBitParam(0.01, "pMutPerBit", "Probability of flipping 1 bit in bit-flip mutation", 'b');
    parser.processParam( pMutPerBitParam, "Genetic Operators" );
    P_MUT_PER_BIT = pMutPerBitParam.value();


    eoValueParam<double> pMut1(0.5, "pMutace1", "Pst stare mutace", 'b');
    parser.processParam( pMut1, "Stara mutace" );
    P_MUT1 = pMut1.value();

    eoValueParam<double> pMut2(0.5, "pMutace2", "Pst nove mutace", 'd');
    parser.processParam( pMut2, "Nova mutace" );
    P_MUT2 = pMut2.value();

    eoValueParam<double> pElita(0.02, "pElita", "Elita (procenta)", 'e');
    parser.processParam( pElita, "Elita");
    elita = pElita.value();


    eoValueParam<string> saveStatsParam("fitness.log", "LogFile","A file to log statistics",'L');
    parser.processParam( saveStatsParam, "Log Statistics File" );
    string saveStatsName = saveStatsParam.value();


    string str_status = parser.ProgramName() + ".status"; // default value
    eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
    parser.processParam( statusParam, "Persistence" );

    
    eoValueParam<string>  inputParam("mice.png", "InputFile","A file with input",'I');
    parser.processParam( inputParam, "Input File" );
    string inputName = inputParam.value();
    
    
   // do the following AFTER ALL PARAMETERS HAVE BEEN PROCESSED
   // i.e. in case you need parameters somewhere else, postpone these
    if (parser.userNeedsHelp())
      {
        parser.printHelp(cout);
        exit(1);
      }
    if (statusParam.value() != "")
      {
	ofstream os(statusParam.value().c_str());
	os << parser;		// and you can use that file as parameter file
      }

    // open and create target picture 
    Picture pic;
    pic.setFileName(inputName); 
    pic.load(); 

    
    //////////////////////////
    //  Random seed
    //////////////////////////
    rng.reseed(SEED);

    /////////////////////////////
    // Fitness function
    ////////////////////////////
    MyFitness eval(pic);


    ////////////////////////////////
    // Initilisation of population
    ////////////////////////////////
    eoPop<PaletteIndividual> pop;
    for (unsigned int igeno=0; igeno<POP_SIZE; igeno++)
      {
	PaletteIndividual v;           // void individual, to be filled
	v.resize(COLORS);
	for (unsigned ivar=0; ivar<COLORS; ivar++)
	  {
	    v[ivar].r = rng.random(256); 
	    v[ivar].g = rng.random(256); 
	    v[ivar].b = rng.random(256); 
	  }
	eval(v);                 // evaluate it
	pop.push_back(v);        // and put it in the population
      }

    // OUTPUT
    pop.sort();
    cout << "Initial Population" << endl;
    cout << pop;
   
 
    /////////////////////////////////////
    // selection and replacement
    ////////////////////////////////////
    eoDetTournamentSelect<PaletteIndividual> selectOne;   // tournament selection 
    // eoProportionalSelect<PaletteIndividual> selectOne;   // roulette-wheel sel. 
    eoSelectPerc<PaletteIndividual> select(selectOne); 
  

    // REPLACE
    eoDeterministicSaDReplacement<PaletteIndividual> replace(elita,1.0-elita);
  
    // CROSSOVER
    PaletteXover xover;

    // MUTATION
    PaletteMut mutation;
  
    //    eoPropCombinedMonOp<PaletteIndividual> mutation(mutation1,P_MUT1);
    // mutation.add(mutation2,P_MUT2); 

    // The operators are  encapsulated into an eoTRansform object
    eoDynSGATransform<PaletteIndividual> transform(xover, &CROSS_RATE, mutation, &MUT_RATE);


    // CHECKPOINT


    // stop after MAX_GEN generations
    //    eoGenContinue<PaletteIndividual> genCont(MAX_GEN);
    // MyCheck mycheck(pic,10);  // (vypisuje fitness, kazdou 10-tou iteraci ulozi obrazek)
    // eoCombinedContinue<PaletteIndividual> cont(genCont);
    // cont.add(mycheck); 
    //
    //    eoCheckPoint<PaletteIndividual> checkpoint(cont);    
    
    myGui gui(&CROSS_RATE,&MUT_RATE,0,100000); // stopping via gui 
    gui.setSrcImage(&pic);
    
    eoCheckPoint<PaletteIndividual> checkpoint(gui);    

    // Create a counter parameter
    eoValueParam<unsigned> generationCounter(0, "Gen.");
    eoIncrementor<unsigned> increment(generationCounter.value());
    checkpoint.add(increment);
  
    eoBestFitnessStat<PaletteIndividual> bestStat;
    eoSecondMomentStats<PaletteIndividual> SecondStat;
    checkpoint.add(bestStat);
    checkpoint.add(SecondStat);
    

    eoFileMonitor fileMonitor(saveStatsName, " ");
    checkpoint.add(fileMonitor);
    fileMonitor.add(generationCounter);
    fileMonitor.add(bestStat);
    fileMonitor.add(SecondStat);
    
  
    /////////////////////////////////////////
    // the GA algorithm
    ////////////////////////////////////////
    eoEasyEA<PaletteIndividual> gga(checkpoint, 
				    eval, 
				    select, 
				    transform, 
				    replace);


    // Apply algo to pop - that's it!
    gga(pop);
  
    // OUTPUT
    pop.sort();
    cout << "FINAL Population\n" << pop[0] << endl;

    // best solution
    Picture dest; 
    dest.setFileName("result.png");
    dest.create( pop.best_element().palette, pic); 
    dest.save();
}


int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
